@@ -15,11 +15,15 @@ if gevent_socketio_found:
 
 import socketio
 import flask
+from flask import json as flask_json
 from werkzeug.debug import DebuggedApplication
 from werkzeug.serving import run_with_reloader
 
+from .namespace import Namespace
 from .test_client import SocketIOTestClient
 
+__version__ = '2.8.1'
+
 
 class _SocketIOMiddleware(socketio.Middleware):
     """This WSGI middleware simply exposes the Flask application in the WSGI
@@ -78,15 +82,20 @@ class SocketIO(object):
     :param json: An alternative json module to use for encoding and decoding
                  packets. Custom json modules must have ``dumps`` and ``loads``
                  functions that are compatible with the standard library
-                 versions.
+                 versions. To use the same json encoder and decoder as a Flask
+                 application, use ``flask.json``.
 
     The Engine.IO server configuration supports the following settings:
 
-    :param async_mode: The library used for asynchronous operations. Valid
-                       options are "threading", "eventlet" and "gevent". If
-                       this argument is not given, "eventlet" is tried first,
-                       then "gevent", and finally "threading". The websocket
-                       transport is not supported in "threading" mode.
+    :param async_mode: The asynchronous model to use. See the Deployment
+                       section in the documentation for a description of the
+                       available options. Valid async modes are
+                       ``threading``, ``eventlet``, ``gevent`` and
+                       ``gevent_uwsgi``. If this argument is not given,
+                       ``eventlet`` is tried first, then ``gevent_uwsgi``,
+                       then ``gevent``, and finally ``threading``. The
+                       first async mode that has all its dependencies installed
+                       is then one that is chosen.
     :param ping_timeout: The time in seconds that the client waits for the
                          server to respond before disconnecting.
     :param ping_interval: The interval in seconds at which the client pings
@@ -112,20 +121,29 @@ class SocketIO(object):
 
     def __init__(self, app=None, **kwargs):
         self.server = None
-        self.server_options = None
+        self.server_options = {}
         self.wsgi_server = None
         self.handlers = []
+        self.namespace_handlers = []
         self.exception_handlers = {}
         self.default_exception_handler = None
-        if app is not None or len(kwargs) > 0:
+        # We can call init_app when:
+        # - we were given the Flask app instance (standard initialization)
+        # - we were not given the app, but we were given a message_queue
+        #   (standard initialization for auxiliary process)
+        # In all other cases we collect the arguments and assume the client
+        # will call init_app from an app factory function.
+        if app is not None or 'message_queue' in kwargs:
             self.init_app(app, **kwargs)
+        else:
+            self.server_options.update(kwargs)
 
     def init_app(self, app, **kwargs):
         if app is not None:
             if not hasattr(app, 'extensions'):
                 app.extensions = {}  # pragma: no cover
             app.extensions['socketio'] = self
-        self.server_options = kwargs
+        self.server_options.update(kwargs)
 
         if 'client_manager' not in self.server_options:
             url = kwargs.pop('message_queue', None)
@@ -140,6 +158,25 @@ class SocketIO(object):
                                     write_only=write_only)
                 self.server_options['client_manager'] = queue
 
+        if 'json' in self.server_options and \
+                self.server_options['json'] == flask_json:
+            # flask's json module is tricky to use because its output
+            # changes when it is invoked inside or outside the app context
+            # so here to prevent any ambiguities we replace it with wrappers
+            # that ensure that the app context is always present
+            class FlaskSafeJSON(object):
+                @staticmethod
+                def dumps(*args, **kwargs):
+                    with app.app_context():
+                        return flask_json.dumps(*args, **kwargs)
+
+                @staticmethod
+                def loads(*args, **kwargs):
+                    with app.app_context():
+                        return flask_json.loads(*args, **kwargs)
+
+            self.server_options['json'] = FlaskSafeJSON
+
         resource = kwargs.pop('path', kwargs.pop('resource', 'socket.io'))
         if resource.startswith('/'):
             resource = resource[1:]
@@ -147,6 +184,9 @@ class SocketIO(object):
         self.async_mode = self.server.async_mode
         for handler in self.handlers:
             self.server.on(handler[0], handler[1], namespace=handler[2])
+        for namespace_handler in self.namespace_handlers:
+            self.server.register_namespace(namespace_handler)
+
         if app is not None:
             # here we attach the SocketIO middlware to the SocketIO object so it
             # can be referenced later if debug middleware needs to be inserted
@@ -177,36 +217,9 @@ class SocketIO(object):
 
         def decorator(handler):
             def _handler(sid, *args):
-                if sid not in self.server.environ:
-                    # we don't have record of this client, ignore this event
-                    return '', 400
-                app = self.server.environ[sid]['flask.app']
-                with app.request_context(self.server.environ[sid]):
-                    if 'saved_session' in self.server.environ[sid]:
-                        self._copy_session(
-                            self.server.environ[sid]['saved_session'],
-                            flask.session)
-                    flask.request.sid = sid
-                    flask.request.namespace = namespace
-                    flask.request.event = {'message': message, 'args': args}
-                    try:
-                        if message == 'connect':
-                            ret = handler()
-                        else:
-                            ret = handler(*args)
-                    except:
-                        err_handler = self.exception_handlers.get(
-                            namespace, self.default_exception_handler)
-                        if err_handler is None:
-                            raise
-                        type, value, traceback = sys.exc_info()
-                        return err_handler(value)
-                    if flask.session.modified and sid in self.server.environ:
-                        self.server.environ[sid]['saved_session'] = {}
-                        self._copy_session(
-                            flask.session,
-                            self.server.environ[sid]['saved_session'])
-                    return ret
+                return self._handle_event(handler, message, namespace, sid,
+                                          *args)
+
             if self.server:
                 self.server.on(message, _handler, namespace=namespace)
             else:
@@ -254,6 +267,40 @@ class SocketIO(object):
         self.default_exception_handler = exception_handler
         return exception_handler
 
+    def on_event(self, message, handler, namespace=None):
+        """Register a SocketIO event handler.
+
+        ``on_event`` is the non-decorator version of ``'on'``.
+
+        Example::
+
+            def on_foo_event(json):
+                print('received json: ' + str(json))
+
+            socketio.on_event('my event', on_foo_event, namespace='/chat')
+
+        :param message: The name of the event. This is normally a user defined
+                        string, but a few event names are already defined. Use
+                        ``'message'`` to define a handler that takes a string
+                        payload, ``'json'`` to define a handler that takes a
+                        JSON blob payload, ``'connect'`` or ``'disconnect'``
+                        to create handlers for connection and disconnection
+                        events.
+        :param handler: The function that handles the event.
+        :param namespace: The namespace on which the handler is to be
+                          registered. Defaults to the global namespace.
+        """
+        self.on(message, namespace=namespace)(handler)
+
+    def on_namespace(self, namespace_handler):
+        if not isinstance(namespace_handler, Namespace):
+            raise ValueError('Not a namespace instance.')
+        namespace_handler._set_socketio(self)
+        if self.server:
+            self.server.register_namespace(namespace_handler)
+        else:
+            self.namespace_handlers.append(namespace_handler)
+
     def emit(self, event, *args, **kwargs):
         """Emit a server generated SocketIO event.
 
@@ -275,23 +322,28 @@ class SocketIO(object):
         :param room: Send the message to all the users in the given room. If
                      this parameter is not included, the event is sent to
                      all connected users.
-        :param include_self: ``True`` to include the sender when broadcasting
-                             or addressing a room, or ``False`` to send to
-                             everyone but the sender.
+        :param skip_sid: The session id of a client to ignore when broadcasting
+                         or addressing a room. This is typically set to the
+                         originator of the message, so that everyone except
+                         that client receive the message.
         :param callback: If given, this function will be called to acknowledge
                          that the client has received the message. The
                          arguments that will be passed to the function are
                          those provided by the client. Callback functions can
                          only be used when addressing an individual client.
         """
-        skip_sid = flask.request.sid \
-            if not kwargs.get('include_self', True) else None
-        self.server.emit(event, *args, namespace=kwargs.get('namespace', '/'),
-                         room=kwargs.get('room'), skip_sid=skip_sid,
-                         callback=kwargs.get('callback'))
+        namespace = kwargs.pop('namespace', '/')
+        room = kwargs.pop('room', None)
+        include_self = kwargs.pop('include_self', True)
+        skip_sid = kwargs.pop('skip_sid', None)
+        if not include_self and not skip_sid:
+            skip_sid = flask.request.sid
+        callback = kwargs.pop('callback', None)
+        self.server.emit(event, *args, namespace=namespace, room=room,
+                         skip_sid=skip_sid, callback=callback, **kwargs)
 
     def send(self, data, json=False, namespace=None, room=None,
-             callback=None, include_self=True):
+             callback=None, include_self=True, skip_sid=None, **kwargs):
         """Send a server-generated SocketIO message.
 
         This function sends a simple SocketIO message to one or more connected
@@ -308,22 +360,23 @@ class SocketIO(object):
         :param room: Send the message only to the users in the given room. If
                      this parameter is not included, the message is sent to
                      all connected users.
-        :param include_self: ``True`` to include the sender when broadcasting
-                             or addressing a room, or ``False`` to send to
-                             everyone but the sender.
+        :param skip_sid: The session id of a client to ignore when broadcasting
+                         or addressing a room. This is typically set to the
+                         originator of the message, so that everyone except
+                         that client receive the message.
         :param callback: If given, this function will be called to acknowledge
                          that the client has received the message. The
                          arguments that will be passed to the function are
                          those provided by the client. Callback functions can
                          only be used when addressing an individual client.
         """
-        skip_sid = flask.request.sid if not include_self else None
+        skip_sid = flask.request.sid if not include_self else skip_sid
         if json:
             self.emit('json', data, namespace=namespace, room=room,
-                      skip_sid=skip_sid, callback=callback)
+                      skip_sid=skip_sid, callback=callback, **kwargs)
         else:
             self.emit('message', data, namespace=namespace, room=room,
-                      skip_sid=skip_sid, callback=callback)
+                      skip_sid=skip_sid, callback=callback, **kwargs)
 
     def close_room(self, room, namespace=None):
         """Close a room.
@@ -414,6 +467,7 @@ class SocketIO(object):
         elif self.server.eio.async_mode == 'eventlet':
             def run_server():
                 import eventlet
+                import eventlet.wsgi
                 eventlet_socket = eventlet.listen((host, port))
 
                 # If provided an SSL argument, use an SSL socket
@@ -514,6 +568,38 @@ class SocketIO(object):
         """Return a simple SocketIO client that can be used for unit tests."""
         return SocketIOTestClient(app, self, namespace)
 
+    def _handle_event(self, handler, message, namespace, sid, *args):
+        if sid not in self.server.environ:
+            # we don't have record of this client, ignore this event
+            return '', 400
+        app = self.server.environ[sid]['flask.app']
+        with app.request_context(self.server.environ[sid]):
+            if 'saved_session' in self.server.environ[sid]:
+                self._copy_session(
+                    self.server.environ[sid]['saved_session'],
+                    flask.session)
+            flask.request.sid = sid
+            flask.request.namespace = namespace
+            flask.request.event = {'message': message, 'args': args}
+            try:
+                if message == 'connect':
+                    ret = handler()
+                else:
+                    ret = handler(*args)
+            except:
+                err_handler = self.exception_handlers.get(
+                    namespace, self.default_exception_handler)
+                if err_handler is None:
+                    raise
+                type, value, traceback = sys.exc_info()
+                return err_handler(value)
+            if flask.session.modified and sid in self.server.environ:
+                self.server.environ[sid]['saved_session'] = {}
+                self._copy_session(
+                    flask.session,
+                    self.server.environ[sid]['saved_session'])
+            return ret
+
     def _copy_session(self, src, dest):
         for k in src:
             dest[k] = src[k]
